/*
 * This is the  STREAMS device driver for NOP.
 *
 * Here we assume there is a NOP device (no-Operation) which
 * has a controller board which interrupts.
 *
 * 1: NOP is just the device driver, not a stream module.
 *
 * 2: Three files:
 *    nopstream.c - This file contains just STREAM routines
 *    nopinit.c   - The initalise and load/unload code
 *    nophrdw.c   - Hardware related code. In particular the interrupts
 */

/*
 * This is where the driver fits in:
 *
 *  +---------------------------+
 *  | STREAM Head               |
 *  +---------------------------+
 *        ^                |
 *        |                v
 *
 *  Other (possible) STREAM modules
 *
 *
 *  Read Side        Write side
 *        ^                |
 *        |                |
 *        |                v
 *  +----------+     +----------+   \
 *  |          |     |          |   |
 *  |          |     |          |   |== We are here
 *  |          |     |          |   |
 *  +----------+     +----------+   /
 *       | (intr)         |
 *  +---------------------------+
 *  | Hardware board            |
 *  +---------------------------+
 *
 *
 * We connect to the rest of the OS at several points:
 *
 *
 * -> nop_load   - Called when driver is dynamicly loaded
 *
 * -> nop_unload - Called when driver is dynamicly unloaded
 *
 * -> nop_open   - Called to open device
 *
 * -> nop_close  - Called to close device
 *
 * -> nopwput    - Called When data is to be written to device
 *
 * -> noprsvr    - Called every so often, when read data can serviced
 *
 * -> nopintr    - Called When device sends interrupt
 */







#include "nop.h"

/* The STREAMS routines:
 *
 * nopopen  - Attached to read side, called when user does open()
 * nopclose - Attached to read side, called when user does close().
 *
 * nopwput  - The Write side put routine, called when user does a write.
 * noprsrv  - The Read  side service routine, scheduled to run AFTER data
 *	      has arrived (often via INTERRUPT routine).
 *
 *
 */
STATIC int nopopen(queue_t		*q,
		   dev_t		*dev,
		   int		flag,
		   int		sflag,
		   struct	cred	*credp);

STATIC void nopclose(queue_t	*q);


STATIC void nopwput(queue_t	*q,
		   mblk_t	*mp);

STATIC void noprsrv(queue_t	*q);


/*
 * Utility Routines, unless used in this file are of no other use.
 *
 * Just here as examples:
 *
 * nopcmds: Do straem commands
 * nopioctl: Do ioctl() requests
 *
 */



STATIC void nopcmds(queue_t	*q,
		   mblk_t	*mp);

STATIC void nopioctl(queue_t	*q,
		    mblk_t	*mp);

typedef int (*PSFI)();	/* Pointer to STREAMS Function return Int */

STATIC  struct	module_info	noprminfo = 
{
  NOP_ID, NOP_NAME, NOP_MIN_PACKET, NOP_MAX_PACKET, NOP_HIWATER, NOP_LOWATER,
};

STATIC  struct	module_info	nopwminfo = 
{
  NOP_ID, NOP_NAME, NOP_MIN_PACKET, NOP_MAX_PACKET, NOP_HIWATER, NOP_LOWATER,
};

STATIC  struct	qinit		noprinit = 
{
  NULL, (PSFI)noprsrv, nopopen, (PSFI)nopclose, NULL, &noprminfo, NULL,
};
STATIC  struct	qinit		nopwinit = 
{
  (PSFI)nopwput, NULL, NULL, NULL, NULL, &nopwminfo, NULL,
};

struct	streamtab	nopinfo = { &noprinit, &nopwinit, NULL, NULL, };


/******************************************************************************
 *  nopopen()
 ******************************************************************************/

/* Frig for nop, should be kmem_alloced array or some such */
STATIC  struct private	sample_private;

/* ARGSUSED */
STATIC int nopopen(q, dev, flag, sflag, credp)
queue_t		*q;
dev_t		*dev;
int		flag;
int		sflag;
struct	cred	*credp;
  {
  int		i, old;

  nop_bdconfig_t	*bd;
  struct private	*private;

  major_t	major = getmajor(*dev);
  minor_t	minor = getminor(*dev);

  DEBUG_ENTER("nopopen", ("q=%x, *dev=%d, sflag=%x, credp=%x", q    , *dev   , sflag   , credp) );


  NOP_LOG(strlog(NOP_ID, 0, 3, SL_TRACE,
		"nopopen - major %d minor %d queue %x",
		(int)major, (int)minor, (int)q));

  /*
   *  Find the board structure for this major number.
   *  (Scan table, defined in Space.c and match on major/minor No)
   */

  for (i = nopboards, bd = nopconfig; i; bd++, i--)
    if (bd->major == major)
      break;
	
  if (i == 0) 
    {
    NOP_LOG(strlog(NOP_ID, 0, 1, SL_TRACE,
		  "nopopen - invalid major number (%d)", (int)major));
    /* return (ENXIO); */
    }

  /* By now we know what peice of hardware is being talked about.
   * Now we must decide if (at init time) we have noticed an actual
   * device sitting there
   */

  /*
   *  Check if we found a board there.
   */
  if (!(bd->flags & 99))  /* Suitable define, set in init code */
    {
    NOP_LOG(strlog(NOP_ID, 0, 1, SL_TRACE,
		  "nopopen - board for major (%d) not installed",
		  (int)major));
    /* return (ENXIO); */
    }

  /* Is the minor No valid? Simple minded test here assumes continuious range
   * of minor Nos.
   */

  if (minor > bd->max_privates) 
    {
    NOP_LOG(strlog(NOP_ID, 0, 1, SL_TRACE,
		  "nopopen - invalid minor number (%d)", (int)minor));
    /* return (ECHRNG); */
    }

  private=&sample_private;	/* Frig for NOP, locate real data structure */
  
  /*
   *  If this stream has not been opened before, set it up.
   */
  
  if (!q->q_ptr) 
    {
    old = splstr();

    private->state   = 0;
    private->read_q  = q;
    private->write_q = WR(q);

    /* On BOTH read and write side, note our private data structure */

    WR(q)->q_ptr = q->q_ptr = (caddr_t) private;

    /*
     *  Need to keep track of priviledge for later reference in 
     *  bind requests.
     */
    if (drv_priv(credp) == 0)
      private->flags = 1;

    splx(old);
    }

  *dev = makedevice(major, minor);

  /* now go ahead and do stuff */

  DEBUG_INT_RETURN (0);
  }


/******************************************************************************
 *  nopclose()
 ******************************************************************************/
STATIC void
nopclose(q)
queue_t	*q;
  {
  struct private	*private = (nop_private_t *)q->q_ptr;
  int		old;

  DEBUG_ENTER("nopclose", ("q=%x",  q ));

  NOP_LOG(strlog(NOP_ID, 1, 3, SL_TRACE, "nopclose queue %x", (int)q));

  /*
   *  Cleanup PRIVATE structure
   */
  old = splstr();

  private->state    = 0;
  private->read_q   = NULL;
  private->write_q  = NULL;
  private->flags    = 0;

  splx(old);

  /* handle any board-specific close processing */
  /* nopbdspecclose(q); */

  DEBUG_VOID_RETURN;
}

/******************************************************************************
 *  nopwput()
 ******************************************************************************/
STATIC void
nopwput(q, mp)
queue_t	*q;
mblk_t	*mp;
{
  DEBUG_ENTER("nopwput", ("q=%x, mp=%x",
			   q   , mp));

  NOP_LOG(strlog(NOP_ID, 2, 3, SL_TRACE,
		"nopwput queue %x type %d", (int)q, mp->b_datap->db_type));

  switch(mp->b_datap->db_type) 
    {
    /*
     *  Process data link commands.
     */
    case M_PROTO:
    case M_PCPROTO:
      DEBUG_TRACE(10, ("M_PROTO/M_PCPROTO"));
      nopcmds(q, mp);
      break;

    /*
     *  Flush read and write queues.
     */
    case M_FLUSH:

      DEBUG_TRACE(10, ("M_FLUSH"));
  
      if (*mp->b_rptr & FLUSHW)
	flushq(q, FLUSHDATA);

      if (*mp->b_rptr & FLUSHR) 
	{
	flushq(RD(q), FLUSHDATA);
	*mp->b_rptr &= ~FLUSHW;
	qreply(q, mp);
	} 
      else
	freemsg(mp);
      break;

    /*
     *  Process ioctls.
     */
    case M_IOCTL:
      DEBUG_TRACE(10, ("M_IOCTL"));
      nopioctl(q, mp);
      break;

    /*
     *  Process data
     */
    case M_DATA:
      DEBUG_TRACE(10, ("M_DATA"));
      nop_m_data(q, mp);  /* Process the message (use hardware) */
      break;
      

    /* 
     *  Anything else we dump.
     */
    default:
      DEBUG_TRACE(10, ("defualt type"));
      NOP_LOG(strlog(NOP_ID, 2, 1, SL_TRACE,
		    "nopwput - unsupported type (%d)",
		    mp->b_datap->db_type));
      freemsg(mp);
      break;
    }
  
  DEBUG_VOID_RETURN;
}

/******************************************************************************
 *  nopcmds()
 ******************************************************************************/
STATIC void
nopcmds(q, mp)
queue_t	*q;
mblk_t	*mp;
  {
  /* LINTED pointer assignment */
  union nop_primitives  *dl = (union nop_primitives *)mp->b_datap->db_base;

  DEBUG_ENTER("nopcmds", ("q=%x,mp=%x", q, mp));

  /* Process the command */

  /*
   *  Free the request.
   */
  freemsg(mp);

  DEBUG_VOID_RETURN;
}

/******************************************************************************
 *  nopioctl()
 ******************************************************************************/
STATIC void
nopioctl(q, mp)
queue_t	*q;
mblk_t	*mp;
  {
  /* LINTED pointer assignment */
  struct iocblk	*ioctl_req = (struct iocblk *)mp->b_rptr;

  nop_private_t	*private       = (nop_private_t *)q->q_ptr;

  short count = 0;

  DEBUG_ENTER("nopioctl", ("q=%x,mp=%x", q, mp));

  NOP_LOG(strlog(NOP_ID, 8, 3, SL_TRACE,
		"nopioctl request for command %d on queue %x",
					ioctl_req->ioc_cmd, (int)q));
  /*
   *  Assume good stuff for now.
   */
  ioctl_req->ioc_error = 0;
  mp->b_datap->db_type = M_IOCACK;

  /*
   *  Screen for priviledged requests.
   */
  switch (ioctl_req->ioc_cmd) 
    {
    /* ioctls that require priviledge */
    }

  /*
   *  Make sure IOCTL's that require buffers have them.
   */
  switch (ioctl_req->ioc_cmd) 
    {
    /* ioctls that require buffers */
    }

  /*
   *  Process request.
   */
  switch (ioctl_req->ioc_cmd) 
    {
    /* do the actual work */
    }

  qreply(q, mp);

  DEBUG_VOID_RETURN;
  }

static int attempts=0 ; /* GPV special */

/******************************************************************************
 *  noprsrv()
 ******************************************************************************/
STATIC void
noprsrv(q)
queue_t	*q;
  {
  nop_private_t	*private;


  mblk_t		*mp_data;
  void *hdr;

  DEBUG_ENTER("nopsrv", ("q=%x", q));

  /*
   *  Process all messages waiting.
   */

  private = (nop_private_t *)q->q_ptr;

  while (mp_data = getq(q)) 
    {
    /*
     *  Set some convient pointers.
     */
    /* LINTED pointer assignment */
    hdr = mp_data->b_rptr;


    if (canput(q->q_next)) 
      {
      attempts=0;
      
      putnext(q, mp_data);
    
      NOP_LOG(strlog(NOP_ID, 9, 3, SL_TRACE, "noprsrv queue %x",
		     (int)q));
      }
    else
      {
      if (++attempts < 100)
	putbq(q, mp_data); /* Reschedule ourselves for later */
      else
	{
	NOP_LOG(strlog(NOP_ID, 9, 3, SL_TRACE, "noprsrv gave up on queue %x",
		     (int)q));
	}
      
      }
    
    }
  

  DEBUG_VOID_RETURN;
}
